package com.ningdatech.pmapi.meeting.manage;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import cn.hutool.crypto.SecureUtil;
import cn.hutool.json.JSONUtil;
import com.alibaba.fastjson.JSON;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.update.LambdaUpdateWrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.ningdatech.basic.exception.BizException;
import com.ningdatech.basic.model.IdVo;
import com.ningdatech.basic.model.PageVo;
import com.ningdatech.basic.util.CollUtils;
import com.ningdatech.basic.util.StrPool;
import com.ningdatech.basic.util.ValidUtil;
import com.ningdatech.cache.lock.DistributedLock;
import com.ningdatech.pmapi.common.helper.RegionCacheHelper;
import com.ningdatech.pmapi.common.model.entity.ViewRegionDTO;
import com.ningdatech.pmapi.expert.constant.ReviewResultEnum;
import com.ningdatech.pmapi.expert.entity.ExpertUserFullInfo;
import com.ningdatech.pmapi.expert.model.entity.ExpertReview;
import com.ningdatech.pmapi.expert.service.IExpertReviewService;
import com.ningdatech.pmapi.expert.service.IExpertUserFullInfoService;
import com.ningdatech.pmapi.meeting.builder.ExpertInviteBuilder;
import com.ningdatech.pmapi.meeting.entity.domain.*;
import com.ningdatech.pmapi.meeting.entity.dto.*;
import com.ningdatech.pmapi.meeting.entity.enumeration.ExpertAttendStatusEnum;
import com.ningdatech.pmapi.meeting.entity.enumeration.ExpertInviteTypeEnum;
import com.ningdatech.pmapi.meeting.entity.enumeration.MeetingReviewTypeEnum;
import com.ningdatech.pmapi.meeting.entity.enumeration.MeetingStatusEnum;
import com.ningdatech.pmapi.meeting.entity.req.*;
import com.ningdatech.pmapi.meeting.entity.vo.*;
import com.ningdatech.pmapi.meeting.helper.ExpertInviteHelper;
import com.ningdatech.pmapi.meeting.helper.MeetingCallOrMsgHelper;
import com.ningdatech.pmapi.meeting.helper.MeetingDataScopeHelper;
import com.ningdatech.pmapi.meeting.helper.MeetingManageHelper;
import com.ningdatech.pmapi.meeting.service.*;
import com.ningdatech.pmapi.meeting.task.ExpertRandomInviteTask;
import com.ningdatech.pmapi.meta.helper.DictionaryCache;
import com.ningdatech.pmapi.meta.helper.TagCache;
import com.ningdatech.pmapi.organization.service.IDingOrganizationService;
import com.ningdatech.pmapi.organization.service.IGovBusinessStripService;
import com.ningdatech.pmapi.projectlib.enumeration.ProjectStatusEnum;
import com.ningdatech.pmapi.projectlib.model.entity.Project;
import com.ningdatech.pmapi.projectlib.service.IProjectService;
import com.ningdatech.pmapi.sys.model.dto.RegionDTO;
import com.ningdatech.pmapi.user.security.auth.model.UserInfoDetails;
import com.ningdatech.pmapi.user.util.LoginUserUtil;
import lombok.RequiredArgsConstructor;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;

import java.time.LocalDateTime;
import java.util.*;
import java.util.function.BiFunction;
import java.util.stream.Collectors;

import static com.ningdatech.pmapi.meeting.entity.enumeration.ExpertAttendStatusEnum.*;
import static com.ningdatech.pmapi.meeting.entity.enumeration.ExpertInviteTypeEnum.APPOINT;
import static com.ningdatech.pmapi.meeting.helper.ExpertInviteHelper.getExpertInviteRule;

/**
 * <p>
 * MeetingManage
 * </p>
 *
 * @author WendyYang
 * @since 13:35 2022/7/26
 */
@Component
@RequiredArgsConstructor
public class MeetingManage {

    private final RegionCacheHelper regionCache;
    private final IMeetingService meetingService;
    private final IExpertInviteAvoidRuleService inviteAvoidRuleService;
    private final IExpertInviteRuleService inviteRuleService;
    private final IExpertUserFullInfoService expertUserFullInfoService;
    private final TagCache tagCache;
    private final DictionaryCache dictionaryCache;
    private final IMeetingExpertService meetingExpertService;
    private final ExpertInviteManage expertInviteManage;
    private final ExpertRandomInviteTask expertRandomInviteTask;
    private final MeetingManageHelper meetingManageHelper;
    private final DistributedLock distributedLock;
    private final IProjectService projectService;
    private final IMeetingInnerProjectService meetingInnerProjectService;
    private final IMeetingOuterProjectService meetingOuterProjectService;
    private final IGovBusinessStripService businessStripService;
    private final IDingOrganizationService dingOrganizationService;
    private final IExpertReviewService expertReviewService;
    private final ExpertInviteHelper expertInviteHelper;
    private final MeetingCallOrMsgHelper meetingCallOrMsgHelper;
    private final MeetingDataScopeHelper meetingDataScopeHelper;
    private static final String INVITED_RULE_CREATE = "INVITED_RULE_CREATE:";

    private static final String MEETING_CREATE_KEY = "MEETING_CREATE:";

    private String getDictName(String dictType) {
        return Optional.ofNullable(dictionaryCache.getByCode(dictType))
                .flatMap(w -> Optional.of(w.getName()))
                .orElse(StrPool.EMPTY);
    }

    /**
     * 锁定时间10 * 60秒
     */
    public static final int RETRY_TIMES = 3;

    @Transactional(rollbackFor = Exception.class)
    public IdVo<Long> meetingCreateAndInviteExpert(MeetingCreateReq req) {
        MeetingBasicDTO meetingBasic = req.getMeetingBasicInfo();
        meetingManageHelper.checkReviewProject(meetingBasic);
        String md5ByParam = SecureUtil.md5(meetingBasic.toString());
        String key = MEETING_CREATE_KEY + md5ByParam;
        if (!distributedLock.lock(key, RETRY_TIMES)) {
            throw BizException.wrap("会议正在创建中");
        }
        try {
            ExpertInviteReq inviteRule = req.getExpertInviteRule();
            // 保存会议基本信息
            Meeting meeting = BeanUtil.copyProperties(meetingBasic, Meeting.class);
            meeting.setStatus(MeetingStatusEnum.NORMAL.getCode());
            UserInfoDetails userDetail = LoginUserUtil.loginUserDetail();
            meeting.setHoldOrg(userDetail.getOrganizationName());
            meeting.setHoldOrgCode(userDetail.getOrganizationCode());
            meeting.setRegionCode(userDetail.getRegionCode());
            meeting.setRegionLevel(userDetail.getRegionLevel());
            meeting.setCreator(userDetail.getUsername());
            meeting.setInviteStatus(Boolean.FALSE);
            meeting.setConfirmedRoster(Boolean.FALSE);
            meeting.setInviteType(inviteRule.getInviteType());
            meetingService.save(meeting);
            if (meetingBasic.getIsInnerProject()) {
                List<MeetingInnerProject> projects = meetingBasic.getInnerProjects().stream().map(w -> {
                    MeetingInnerProject project = BeanUtil.copyProperties(w, MeetingInnerProject.class);
                    project.setMeetingId(meeting.getId());
                    return project;
                }).collect(Collectors.toList());
                meetingInnerProjectService.saveBatch(projects);
            } else {
                List<MeetingOuterProject> projects = meetingBasic.getProjects().stream().map(w -> {
                    MeetingOuterProject project = BeanUtil.copyProperties(w, MeetingOuterProject.class);
                    project.setMeetingId(meeting.getId());
                    return project;
                }).collect(Collectors.toList());
                meetingOuterProjectService.saveBatch(projects);
            }
            // 抽取专家
            req.getExpertInviteRule().setMeetingId(meeting.getId());
            expertInviteByCreate(inviteRule);
            return IdVo.of(meeting.getId());
        } finally {
            distributedLock.releaseLock(key);
        }
    }

    @Transactional(rollbackFor = Exception.class)
    public void continueInvite(Long meetingId) {
        String key = "CONTINUE_INVITE:" + meetingId;
        if (!distributedLock.lock(key, RETRY_TIMES)) {
            throw BizException.wrap("已进行续抽，请勿重复点击");
        }
        try {
            Meeting meeting = meetingService.getById(meetingId);
            if (!meeting.getInviteStatus()) {
                throw BizException.wrap("该会议正在抽取专家，暂无法续抽");
            }
            if (meeting.getStartTime().isBefore(LocalDateTime.now())) {
                throw BizException.wrap("会议已开始，不允许续抽");
            }
            if (!MeetingStatusEnum.NORMAL.eq(meeting.getStatus())) {
                throw BizException.wrap("续抽失败，请刷新后重试");
            }
            boolean invitedContinue = meetingManageHelper.checkCouldBeInvitedContinue(meetingId);
            if (!invitedContinue) {
                throw BizException.wrap("抽取人员数量已满足抽取规则");
            }
            expertRandomInviteTask.notifyInviteTask(meetingId);
        } finally {
            distributedLock.releaseLock(key);
        }
    }

    @Transactional(rollbackFor = Exception.class)
    public void convertToAppoint(Long meetingId) {
        String key = "CONVERT_TO_APPOINT:" + meetingId;
        if (!distributedLock.lock(key, RETRY_TIMES)) {
            throw BizException.wrap("已进行转换，请勿重复点击");
        }
        try {
            Meeting meeting = meetingService.getById(meetingId);
            if (!MeetingStatusEnum.NORMAL.eq(meeting.getStatus())) {
                throw BizException.wrap("转换失败，请刷新后重试");
            }
            expertRandomInviteTask.cancelByMeetingId(meetingId);
            LambdaUpdateWrapper<Meeting> meetingUpdate = Wrappers.lambdaUpdate(Meeting.class)
                    .set(Meeting::getInviteType, APPOINT.getCode())
                    .eq(Meeting::getId, meetingId);
            meetingService.update(meetingUpdate);
            saveAppointRuleByConvertFromRandomRule(meetingId);
        } finally {
            distributedLock.releaseLock(key);
        }
    }

    private void saveAppointRuleByConvertFromRandomRule(Long meetingId) {
        AppointInviteRuleDTO rule = new AppointInviteRuleDTO();
        rule.setInviteDesc("转为指定抽取");
        rule.setExpertIdList(Collections.emptyList());
        rule.setInviteType(APPOINT.getCode());
        rule.setCount(0);
        ExpertInviteRule inviteRule = new ExpertInviteRule();
        inviteRule.setMeetingId(meetingId);
        inviteRule.setInviteType(rule.getInviteType());
        inviteRule.setInviteCount(0);
        inviteRule.setInviteRule(JSONUtil.toJsonStr(rule));
        inviteRuleService.save(inviteRule);
    }

    @Transactional(rollbackFor = Exception.class)
    public void expertInviteByCreate(ExpertInviteReq req) {
        String key = INVITED_RULE_CREATE + req.getMeetingId();
        if (!distributedLock.lock(key, RETRY_TIMES)) {
            throw BizException.wrap("不可重复进行专家抽取");
        }
        try {
            Meeting meeting = meetingService.getById(req.getMeetingId());
            if (ExpertInviteTypeEnum.RANDOM.eq(req.getInviteType())) {
                List<RandomInviteRuleDTO> randomRules = req.getRandomRules();
                Assert.notEmpty(randomRules, "随机抽取规则不能为空");
                AvoidRuleDTO avoidInfo = req.getAvoidRule();
                Assert.notNull(avoidInfo, "回避信息不能为空");
                // 随机抽取的话则需进行抽取数量校验
                LocalDateTime now = LocalDateTime.now();
                expertInviteManage.expertInviteByMeetingCreate(meeting, randomRules, avoidInfo);
                expertRandomInviteTask.addInviteTaskByMeetingCreate(meeting.getId(), now);
                LambdaUpdateWrapper<Meeting> mUpdate = Wrappers.lambdaUpdate(Meeting.class)
                        .set(Meeting::getInviteStatus, false)
                        .eq(Meeting::getId, meeting.getId());
                meetingService.update(mUpdate);
                // 回避规则
                ExpertInviteAvoidRule avoidRule = new ExpertInviteAvoidRule();
                avoidRule.setMeetingId(meeting.getId());
                // 未传值时设置为0 表示不限制周参与次数
                avoidRule.setWeekInviteCount(ObjectUtil.defaultIfNull(avoidRule.getWeekInviteCount(), 0));
                avoidRule.setAvoidOrgIds(CollUtils.joinByComma(avoidInfo.getAvoidOrgIdList()));
                avoidRule.setAvoidUnitIds(CollUtils.joinByComma(avoidInfo.getAvoidUnitIdList()));
                avoidRule.setAvoidExpertIds(CollUtils.joinByComma(avoidInfo.getExpertIds()));
                inviteAvoidRuleService.save(avoidRule);
            } else {
                // 指定邀请
                AppointInviteRuleDTO appointRule = req.getAppointRule();
                Assert.notNull(appointRule, "指定邀请规则不能为空");
                Map<Long, ExpertUserFullInfo> usersMap = expertInviteHelper.checkAppointExpert(appointRule);
                // 处理指定抽取规则
                ExpertInviteRule appoint = getExpertInviteRule(appointRule, meeting.getId());
                inviteRuleService.save(appoint);
                Long ruleId = appoint.getId();
                List<MeetingExpert> experts = appointRule.getExpertIdList().stream().map(w -> {
                    ExpertUserFullInfo info = usersMap.get(w);
                    MeetingExpert expert = ExpertInviteBuilder.getExpertByAppoint(meeting.getId(), info, ruleId);
                    expert.setStatus(NOTICING.getCode());
                    return expert;
                }).collect(Collectors.toList());
                meetingExpertService.saveBatch(experts);
            }
        } finally {
            distributedLock.releaseLock(key);
        }
    }

    /**
     * 抽取数量校验
     *
     * @param req 抽取参数
     * @return {@link ExpertCountOnChangeVO}
     * @author WendyYang
     **/
    public ExpertCountOnChangeVO expertCountOnChange(ExpertInviteReq req) {
        // 未传递指定邀请专家直接忽略
        expertInviteHelper.checkAppointExpert(req.getAppointRule());
        Meeting meeting = meetingService.getById(req.getMeetingId());
        ExpertCountOnChangeVO resultCount = new ExpertCountOnChangeVO();
        resultCount.setStatus(true);
        resultCount.setCountList(new ArrayList<>());
        // 参数校验
        String errMsg = ValidUtil.validFast(req, AbstractInviteRule.CountCheck.class);
        if (errMsg != null) {
            resultCount.setStatus(false);
            resultCount.setMessage(errMsg);
            return resultCount;
        }
        List<Long> expertIdsChoose = new ArrayList<>();
        for (int i = 0; i < req.getRandomRules().size(); i++) {
            RandomInviteRuleDTO randomRule = req.getRandomRules().get(i);
            ExpertChooseDTO chooseExpert = expertInviteManage.expertInviteByRandomRule(req.getAvoidRule(),
                    randomRule, expertIdsChoose, meeting.getStartTime(), meeting.getEndTime());
            if (i < req.getRandomRules().size() - 1 && chooseExpert.getTotal() > 0) {
                List<Long> tempUserIds = CollUtils.fieldList(chooseExpert.getExperts(), ExpertUserFullInfo::getUserId);
                expertIdsChoose.addAll(tempUserIds);
            }
            resultCount.addCountList(chooseExpert.getTotal());
        }
        return resultCount;
    }

    /**
     * 专家会议列表
     *
     * @param req 查询参数
     * @return 会议列表
     * @author WendyYang
     **/
    public PageVo<MeetingByManagerVO> meetingListForExpert(MeetingListReq req) {
        Long expertId = req.getExpertId() != null ? req.getExpertId() : LoginUserUtil.getUserId();
        List<MeetingAndAttendStatusDTO> meetings = meetingExpertService.listByExpertIdAndStatus(expertId, null, null);
        if (meetings.isEmpty()) {
            return PageVo.empty();
        }
        Map<Long, MeetingAndAttendStatusDTO> mapByMeetingId = new HashMap<>(16);
        meetings.forEach(w -> mapByMeetingId.put(w.getMeetingId(), w));
        LambdaQueryWrapper<Meeting> query = new LambdaQueryWrapper<Meeting>()
                .orderByDesc(Meeting::getCreateOn)
                .in(Meeting::getId, mapByMeetingId.keySet())
                .eq(Meeting::getConfirmedRoster, Boolean.TRUE)
                .ne(Meeting::getStatus, MeetingStatusEnum.CANCELED.getCode());
        if (req.getExpertId() == null) {
            meetingManageHelper.buildMeetingQuery(query, req);
        }
        Page<Meeting> page = meetingService.page(req.page(), query);
        if (page.getTotal() == 0) {
            return PageVo.empty();
        }
        PageVo<MeetingByManagerVO> result = new PageVo<>(new ArrayList<>(), page.getTotal());
        page.getRecords().forEach(meeting -> {
            MeetingByManagerVO item = meetingManageHelper.buildByMeeting(meeting);
            MeetingAndAttendStatusDTO info = mapByMeetingId.get(meeting.getId());
            item.setExpertStatus(info.getStatus());
            result.getRecords().add(item);
        });
        return result;
    }

    public PageVo<MeetingByManagerVO> meetingListForLeave(MeetingLeaveListReq req) {
        Long expertId = ObjectUtil.defaultIfNull(req.getExpertId(), LoginUserUtil.getUserId());
        LambdaQueryWrapper<MeetingExpert> meQuery = Wrappers.lambdaQuery(MeetingExpert.class)
                .eq(MeetingExpert::getExpertId, expertId)
                .in(MeetingExpert::getStatus, AGREED.getCode(), ON_LEAVE.getCode());
        List<MeetingExpert> meetings = meetingExpertService.list(meQuery);
        if (meetings.isEmpty()) {
            return PageVo.empty();
        }
        Map<Long, MeetingExpert> meetingMap = new HashMap<>(16);
        meetings.forEach(w -> meetingMap.put(w.getMeetingId(), w));
        LambdaQueryWrapper<Meeting> query = new LambdaQueryWrapper<Meeting>()
                .orderByAsc(Meeting::getStartTime)
                .in(Meeting::getId, meetingMap.keySet())
                .eq(Meeting::getConfirmedRoster, Boolean.TRUE)
                .ne(Meeting::getStatus, MeetingStatusEnum.CANCELED.getCode());
        Page<Meeting> page = meetingService.page(req.page(), query);
        if (page.getTotal() == 0) {
            return PageVo.empty();
        }
        PageVo<MeetingByManagerVO> result = new PageVo<>(new ArrayList<>(), page.getTotal());
        page.getRecords().forEach(meeting -> {
            MeetingByManagerVO item = meetingManageHelper.buildByMeeting(meeting);
            MeetingExpert info = meetingMap.get(meeting.getId());
            item.setExpertStatus(info.getStatus());
            result.getRecords().add(item);
        });
        return result;
    }


    /**
     * 管理员会议列表
     *
     * @param req 查询参数
     * @return 会议列表
     * @author WendyYang
     */
    public PageVo<MeetingByManagerVO> meetingListForManager(MeetingListReq req) {
        ViewRegionDTO viewRegions = meetingDataScopeHelper.meetingListViewRegions();
        Collection<String> regions;
        if ((regions = viewRegions.getRegions()) != null && regions.isEmpty()) {
            return PageVo.empty();
        }
        LambdaQueryWrapper<Meeting> query = new LambdaQueryWrapper<Meeting>()
                .orderByDesc(Meeting::getCreateOn);
        // 补充逻辑 如果拥有超级管理员权限可以看到所有事务
        if (regions != null) {
            query.in(Meeting::getRegionCode, regions);
            query.ge(Meeting::getRegionLevel, viewRegions.getRegionLevel());
        }
        query.eq(req.getStatus() != null, Meeting::getStatus, req.getStatus());
        meetingManageHelper.buildMeetingQuery(query, req);
        Page<Meeting> page = meetingService.page(req.page(), query);
        if (page.getTotal() == 0) {
            return PageVo.empty();
        }
        PageVo<MeetingByManagerVO> result = new PageVo<>(new ArrayList<>(), page.getTotal());
        page.getRecords().forEach(meeting -> {
            MeetingByManagerVO item = meetingManageHelper.buildByMeeting(meeting);
            result.getRecords().add(item);
        });
        return result;
    }

    public MeetingDetailBasicVO getMeetingDetail(Long meetingId) {
        Meeting meeting = meetingService.getById(meetingId);
        Assert.notNull(meeting, "会议不存在");
        MeetingDetailBasicVO detail = MeetingDetailBasicVO.builder()
                .meetingId(meeting.getId())
                .regionCode(meeting.getRegionCode())
                .meetingName(meeting.getName())
                .meetingType(meeting.getType())
                .meetingAddress(meeting.getMeetingAddress())
                .typeName(getDictName(meeting.getType()))
                .startTime(meeting.getStartTime())
                .endTime(meeting.getEndTime())
                .judgesAttendanceTime(meeting.getJudgesAttendanceTime())
                .meetingUsageTime(meeting.getMeetingUsageTime())
                .contact(meeting.getContact())
                .connecter(meeting.getConnecter())
                .status(meeting.getStatus())
                .holdOrg(meeting.getHoldOrg())
                .createOn(meeting.getCreateOn())
                .creator(meeting.getCreator())
                .createBy(meeting.getCreator())
                .inviteType(meeting.getInviteType())
                .confirmedRoster(meeting.getConfirmedRoster())
                .invitedStopped(meeting.getInviteStatus())
                .build();
        if (meeting.getIsInnerProject()) {
            List<MeetingInnerProject> innerProjects = meetingInnerProjectService.listByMeetingId(meetingId);
            List<Project> projects = projectService.listByIds(CollUtils.fieldList(innerProjects, MeetingInnerProject::getProjectId));
            Map<Long, ExpertReview> reviewMap = expertReviewService.listFinalReviewMap(meetingId);
            List<MeetingReviewProjectDTO> convert = CollUtils.convert(projects, w -> {
                MeetingReviewProjectDTO mrp = new MeetingReviewProjectDTO();
                mrp.setProjectId(w.getId());
                mrp.setBuildOrg(w.getBuildOrgName());
                mrp.setProjectName(w.getProjectName());
                mrp.setProjectType(w.getProjectType().toString());
                mrp.setProjectYear(w.getProjectYear());
                mrp.setDeclareAmount(w.getDeclareAmount());
                mrp.setProjectCode(w.getProjectCode());
                ExpertReview review = reviewMap.get(w.getId());
                if (review != null) {
                    mrp.setReviewResult(ReviewResultEnum.getByCode(review.getReviewResult()).getValue());
                }
                return mrp;
            });
            detail.setProjects(convert);
        } else {
            List<MeetingOuterProject> outerProjects = meetingOuterProjectService.listByMeetingId(meetingId);
            detail.setProjects(BeanUtil.copyToList(outerProjects, MeetingReviewProjectDTO.class));
        }
        detail.setInviteRule(inviteRuleDetail(meetingId));
        return detail;
    }

    public ExpertInviteDetailVO inviteExpertList(Long meetingId) {
        Meeting meeting = meetingService.getById(meetingId);
        if (meeting == null) {
            throw BizException.wrap("该会议信息不存在");
        }
        ExpertInviteDetailVO result = new ExpertInviteDetailVO();
        result.setInvitedStopped(meeting.getInviteStatus());
        result.setConfirmedRoster(meeting.getConfirmedRoster());
        List<MeetingExpert> experts = meetingExpertService.listByMeetingId(meetingId);
        if (experts.isEmpty()) {
            return result;
        }
        List<Long> expertIds = CollUtils.fieldList(experts, MeetingExpert::getExpertId);
        Map<Long, ExpertBasicInfoVO> expertMap = meetingManageHelper.getExpertBasicInfo(expertIds);
        BiFunction<MeetingExpert, Map<Long, ExpertInviteRule>, InviteExpertListItemVO> mapping = (me, ruleMap) -> {
            ExpertBasicInfoVO expert = expertMap.get(me.getExpertId());
            InviteExpertListItemVO item = BeanUtil.copyProperties(expert, InviteExpertListItemVO.class);
            item.setExpertMeetingId(me.getId());
            item.setMeetingId(me.getMeetingId());
            item.setStatus(me.getStatus());
            item.setMobile(me.getMobile());
            item.setNoticeTime(me.getCreateOn());
            item.setRuleId(me.getRuleId());
            item.setIsHeadman(me.getIsHeadman());
            ExpertInviteRule rule = ruleMap.get(me.getRuleId());
            item.setInviteType(rule == null ? APPOINT.getCode() : rule.getInviteType());
            if (NOTICING.eq(me.getStatus())) {
                item.setNoticeStatus("通知中");
            } else {
                item.setNoticeStatus("已通知");
            }
            return item;
        };
        List<ExpertInviteRule> inviteRules = inviteRuleService.listByMeetingId(meetingId);
        Map<Long, ExpertInviteRule> ruleMap = CollUtils.listToMap(inviteRules, ExpertInviteRule::getId);
        boolean isRandom = ExpertInviteTypeEnum.RANDOM.eq(meeting.getInviteType());
        if (isRandom) {
            result.setInviteStatistics(new ArrayList<>());
            Map<Long, List<MeetingExpert>> groupByRule = CollUtils.group(experts, MeetingExpert::getRuleId);
            ruleMap.forEach((k, v) -> {
                InviteStatisticsByRuleVO statistics = InviteStatisticsByRuleVO.init(k);
                statistics.setInviteCnt(v.getInviteCount());
                List<MeetingExpert> expertList = groupByRule.get(k);
                if (expertList != null) {
                    expertList.forEach((expert) -> {
                        if (AGREED.eq(expert.getStatus())) {
                            statistics.incrAgreeCnt();
                        }
                        statistics.incrNoticedCnt();
                    });
                }
                result.getInviteStatistics().add(statistics);
            });
        }
        List<InviteExpertListItemVO> converts = CollUtils.convert(experts, me -> mapping.apply(me, ruleMap));
        result.setInviteExpertList(converts);
        // 确定参加列表
        return result;
    }

    /**
     * 会议基本信息修改
     *
     * @param po 修改参数
     **/
    public void meetingBasicInfoModify(MeetingBasicInfoModifyReq po) {
        Meeting meeting = new Meeting();
        BeanUtil.copyProperties(po, meeting);
        LocalDateTime now = LocalDateTime.now();
        Meeting old = meetingService.getById(po.getId());
        List<MeetingExpert> meList = meetingExpertService.listAgreedExperts(Collections.singletonList(po.getId()));
        if (!meList.isEmpty() && old.getStartTime().isAfter(now)) {
            // TODO
            /*String meetingType = dictionaryCache.getByCode(old.getType()).getName();
              List<SendSmsContext> contexts = YxtSmsContextBuilder.smsToExpertByMeetingChange(old, meeting, meList, meetingType);
              yxtCallOrSmsHelper.sendSms(contexts);*/
        }
        meetingService.updateById(meeting);
    }

    /**
     * 抽取规则详情
     *
     * @param meetingId 会议ID
     * @return {@link InviteRuleDetailVO}
     * @author WendyYang
     **/
    public InviteRuleDetailVO inviteRuleDetail(Long meetingId) {
        InviteRuleDetailVO result = new InviteRuleDetailVO();
        List<ExpertInviteRule> inviteRules = inviteRuleService.listByMeetingId(meetingId);
        result.setRandomRules(new ArrayList<>());
        Map<ExpertInviteTypeEnum, List<ExpertInviteRule>> groupByType = CollUtils.group(inviteRules, w -> ExpertInviteTypeEnum.getByCode(w.getInviteType()));
        List<ExpertInviteRule> randoms = groupByType.get(ExpertInviteTypeEnum.RANDOM);
        if (CollUtil.isNotEmpty(randoms)) {
            randoms.forEach(random -> {
                RandomInviteRuleVO randomRule = JSON.parseObject(random.getInviteRule(), RandomInviteRuleVO.class);
                randomRule.setId(random.getId());
                if (randomRule.getExpertTags() != null) {
                    randomRule.getExpertTags().forEach(w -> {
                        List<String> tagNames = CollUtils.convert(w.getTagCodes(), tagCode -> tagCache.getByTagCode(tagCode).getTagName());
                        w.setTagCodes(tagNames);
                    });
                }
                if (randomRule.getExpertDicts() != null) {
                    randomRule.getExpertDicts().forEach(w -> {
                        List<String> dictNames = CollUtils.convert(w.getDictCodes(), dictCode -> dictionaryCache.getByCode(dictCode).getName());
                        w.setDictCodes(dictNames);
                    });
                }
                if (StrUtil.isNotEmpty(randomRule.getIntentionRegionCode())) {
                    List<RegionDTO> intentionRegions = regionCache.listParents(randomRule.getIntentionRegionCode(), randomRule.getIntentionRegionLevel());
                    randomRule.setIntentionRegions(intentionRegions);
                }
                if (StrUtil.isNotEmpty(randomRule.getExpertRegionCode())) {
                    List<RegionDTO> expertRegions = regionCache.listParents(randomRule.getExpertRegionCode(), randomRule.getExpertRegionLevel());
                    randomRule.setExpertRegions(expertRegions);
                }
                result.getRandomRules().add(randomRule);
            });
            AvoidRuleDTO avoidInfo = inviteAvoidRuleService.getAvoidInfo(meetingId);
            AvoidInfoVO vo = new AvoidInfoVO();
            vo.setWeekInviteCount(avoidInfo.getWeekInviteCount());
            vo.setAvoidOrgIds(avoidInfo.getAvoidOrgIdList());
            vo.setAvoidUnitIds(avoidInfo.getAvoidUnitIdList());
            if (CollUtil.isNotEmpty(vo.getAvoidOrgIds())) {
                vo.setAvoidOrgs(businessStripService.listNameByCodes(avoidInfo.getAvoidOrgIdList()));
            }
            if (CollUtil.isNotEmpty(vo.getAvoidUnitIds())) {
                vo.setAvoidUnits(dingOrganizationService.listNameByCodes(avoidInfo.getAvoidUnitIdList()));
            }
            if (CollUtil.isNotEmpty(avoidInfo.getExpertIds())) {
                vo.setExperts(new ArrayList<>(meetingManageHelper.getExpertBasicInfo(avoidInfo.getExpertIds()).values()));
            }
            result.setAvoidInfo(vo);
        } else {
            List<ExpertInviteRule> appoints = groupByType.get(APPOINT);
            ExpertInviteRule appoint = appoints.get(0);
            AppointInviteRuleDTO appointRule = JSON.parseObject(appoint.getInviteRule(), AppointInviteRuleDTO.class);
            appointRule.setId(appoint.getId());
            AppointRuleVO vo = new AppointRuleVO();
            vo.setInviteDesc(appointRule.getInviteDesc());
            vo.setExperts(new ArrayList<>(meetingManageHelper.getExpertBasicInfo(appointRule.getExpertIdList()).values()));
            result.setAppointRule(vo);
        }
        return result;
    }

    public void batchAppointExperts(BatchAppointExpertsReq req) {
        Long meetingId = req.getMeetingId();
        String key = "BATCH_APPOINT_EXPERT:" + meetingId;
        if (!distributedLock.lock(key, RETRY_TIMES)) {
            throw BizException.wrap("补充专家中，请勿重复点击");
        }
        try {
            Meeting meeting = meetingService.getById(meetingId);
            if (!APPOINT.eq(meeting.getInviteType())) {
                throw BizException.wrap("该会议不能指定邀请专家");
            }
            if (meeting.getConfirmedRoster()) {
                throw BizException.wrap("补充专家失败，已确认过专家名单");
            }
            if (!MeetingStatusEnum.NORMAL.eq(meeting.getStatus())) {
                throw BizException.wrap("补充专家失败");
            }
            AppointInviteRuleDTO rule = inviteRuleService.appointRuleByMeetingId(meetingId);
            List<ExpertUserFullInfo> userInfos = meetingManageHelper.appointExpertCheck(meetingId, meeting.getInviteType(), req.getExpertIdList());
            List<MeetingExpert> expertList = CollUtils.convert(userInfos, w -> {
                MeetingExpert me = ExpertInviteBuilder.getExpertByAppoint(meetingId, w, rule.getId());
                me.setStatus(NOTICING.getCode());
                return me;
            });
            meetingExpertService.saveBatch(expertList);
        } finally {
            distributedLock.releaseLock(key);
        }
    }

    public void stopRandomInvite(Long meetingId) {
        expertRandomInviteTask.cancelByMeetingId(meetingId);
    }

    @Transactional(rollbackFor = Exception.class)
    public void cancelMeeting(MeetingCancelReq req) {
        Long meetingId = req.getMeetingId();
        String key = "CANCEL_MEETING:" + meetingId;
        if (!distributedLock.lock(key, RETRY_TIMES)) {
            throw BizException.wrap("正在取消会议，请刷新后重试");
        }
        try {
            Meeting meeting = meetingService.getById(meetingId);
            if (MeetingStatusEnum.CANCELED.eq(meeting.getStatus())) {
                throw BizException.wrap("会议已取消");
            }
            if (meeting.getStartTime().isBefore(LocalDateTime.now())) {
                throw BizException.wrap("会议已开始，暂时无法取消");
            }
            LambdaUpdateWrapper<Meeting> meetingUpdate = Wrappers.lambdaUpdate(Meeting.class)
                    .set(Meeting::getStatus, MeetingStatusEnum.CANCELED.getCode())
                    .eq(Meeting::getId, meetingId);
            meetingService.update(meetingUpdate);
            expertRandomInviteTask.cancelByMeetingId(meetingId);
            // 发送通知给专家
            List<MeetingExpert> experts = meetingExpertService.listAgreedExperts(meetingId);
            if (!experts.isEmpty()) {
                meetingCallOrMsgHelper.sendCancelMeetingMsg(experts, meeting);
            }
        } finally {
            distributedLock.releaseLock(key);
        }
    }

    public ExpertInvitationDetailVO expertInvitationDetail(Long meetingId, Long expertId) {
        Long userId = expertId == null ? LoginUserUtil.getUserId() : expertId;
        MeetingExpert me = meetingExpertService.getByMeetingIdAndExpertId(meetingId, userId);
        if (me == null) {
            throw BizException.wrap("未被邀请参加此会议");
        }
        if (AGREED.eq(me.getStatus())) {
            throw BizException.wrap("未确认参加此会议");
        }
        ExpertUserFullInfo userInfo = expertUserFullInfoService.getByUserId(userId);
        Meeting meeting = meetingService.getById(meetingId);
        return ExpertInvitationDetailVO.builder()
                .expertName(userInfo.getExpertName())
                .holdOrg(meeting.getHoldOrg())
                .meetingName(meeting.getName())
                .startTime(meeting.getStartTime())
                .endTime(meeting.getEndTime())
                .inviteTime(me.getCreateOn())
                .connecter(meeting.getConnecter())
                .contact(meeting.getContact())
                .build();
    }

    public void confirmAttendByManager(ExpertConfirmReq req) {
        String key = "CONFIRM_ATTEND:" + req.getExpertMeetingId();
        String msgPrefix = req.getAgreed() ? "确认参加" : "拒绝参加";
        if (!distributedLock.lock(key, RETRY_TIMES)) {
            throw BizException.wrap("%s失败，请重试！", msgPrefix);
        }
        try {
            MeetingExpert me = meetingExpertService.getById(req.getExpertMeetingId());
            if (!NOTICING.eq(me.getStatus())) {
                throw BizException.wrap("%s失败，请重试！", msgPrefix);
            }
            LambdaUpdateWrapper<MeetingExpert> update = Wrappers.lambdaUpdate(MeetingExpert.class)
                    .set(MeetingExpert::getStatus, (req.getAgreed() ? AGREED : REFUSED).getCode())
                    .eq(MeetingExpert::getId, req.getExpertMeetingId());
            meetingExpertService.update(update);
        } finally {
            distributedLock.releaseLock(key);
        }
    }

    @Transactional(rollbackFor = Exception.class)
    public void expertRemove(ExpertRemoveReq req) {
        String key = "EXPERT_REMOVE:" + req.getExpertMeetingId();
        if (!distributedLock.lock(key, RETRY_TIMES)) {
            throw BizException.wrap("删除专家失败，请重试！");
        }
        try {
            Meeting meeting = meetingService.getById(req.getMeetingId());
            if (MeetingStatusEnum.CANCELED.eq(meeting.getStatus())) {
                throw BizException.wrap("会议已取消！");
            }
            if (LocalDateTime.now().isAfter(meeting.getStartTime())) {
                throw BizException.wrap("会议已开始，不允许移除专家！");
            }
            MeetingExpert expert = meetingExpertService.getById(req.getExpertMeetingId());
            if (!APPOINT.eq(expert.getInviteType())) {
                throw BizException.wrap("随机抽取的专家不允许移除！");
            }
            if (!NOTICING.eq(expert.getStatus())) {
                throw BizException.wrap("已确认过的专家不允许移除！");
            }
            LambdaUpdateWrapper<Meeting> mUpdate = Wrappers.lambdaUpdate(Meeting.class)
                    .set(Meeting::getConfirmedRoster, false)
                    .eq(Meeting::getId, req.getMeetingId());
            meetingService.update(mUpdate);
            meetingExpertService.removeById(req.getExpertMeetingId());
        } finally {
            distributedLock.releaseLock(key);
        }
    }

    public void releaseExperts(MeetingCancelReq req) {
        String key = "EXPERT_RELEASE:" + req.getMeetingId();
        if (!distributedLock.lock(key, RETRY_TIMES)) {
            throw BizException.wrap("释放专家失败，请重试！");
        }
        try {
            Meeting meeting = meetingService.getById(req.getMeetingId());
            if (MeetingStatusEnum.CANCELED.eq(meeting.getStatus())) {
                throw BizException.wrap("会议已取消");
            }
            LambdaUpdateWrapper<MeetingExpert> update = Wrappers.lambdaUpdate(MeetingExpert.class)
                    .set(MeetingExpert::getStatus, RELEASED.getCode())
                    .eq(MeetingExpert::getMeetingId, req.getMeetingId());
            meetingExpertService.update(update);
        } finally {
            distributedLock.releaseLock(key);
        }
    }

    @Transactional(rollbackFor = Exception.class)
    public void setUpHeadman(ExpertConfirmReq req) {
        String key = "SETUP_HEADMAN:" + req.getMeetingId();
        if (!distributedLock.lock(key, RETRY_TIMES)) {
            throw BizException.wrap("设置专家组长失败，请重试！");
        }
        try {
            Meeting meeting = meetingService.getById(req.getMeetingId());
            if (MeetingStatusEnum.CANCELED.eq(meeting.getStatus())) {
                throw BizException.wrap("会议已取消");
            }
            LambdaUpdateWrapper<MeetingExpert> cancel = Wrappers.lambdaUpdate(MeetingExpert.class)
                    .set(MeetingExpert::getIsHeadman, Boolean.FALSE)
                    .eq(MeetingExpert::getIsHeadman, Boolean.TRUE)
                    .eq(MeetingExpert::getMeetingId, req.getMeetingId());
            meetingExpertService.update(cancel);
            MeetingExpert headman = meetingExpertService.getById(req.getExpertMeetingId());
            if (!AGREED.eq(headman.getStatus())) {
                ExpertAttendStatusEnum status = getByCode(headman.getStatus());
                throw BizException.wrap("该专家处于：%s状态，不能被设置为专家组长！", status.getValue());
            }
            LambdaUpdateWrapper<MeetingExpert> setup = Wrappers.lambdaUpdate(MeetingExpert.class)
                    .set(MeetingExpert::getIsHeadman, Boolean.TRUE)
                    .eq(MeetingExpert::getId, req.getExpertMeetingId());
            meetingExpertService.update(setup);
        } finally {
            distributedLock.releaseLock(key);
        }
    }

    public void confirmedRoster(ConfirmedRosterReq req) {
        Long meetingId = req.getMeetingId();
        String key = "MEETING_RESEND_SMS:" + meetingId;
        if (!distributedLock.lock(key, RETRY_TIMES)) {
            throw BizException.wrap("请刷新后重试！");
        }
        try {
            Meeting meeting = meetingService.getById(meetingId);
            if (!MeetingStatusEnum.NORMAL.eq(meeting.getStatus())) {
                throw BizException.wrap("请刷新后重试！");
            }
            if (!meeting.getConfirmedRoster()) {
                // 首次确认的
                LambdaUpdateWrapper<Meeting> update = Wrappers.lambdaUpdate(Meeting.class);
                update.set(Meeting::getConfirmedRoster, Boolean.TRUE);
                update.eq(Meeting::getId, meetingId);
                meetingService.update(update);
            }
            List<MeetingExpert> experts = meetingExpertService.listAgreedExperts(meetingId);
            List<MeetingExpert> expertNoticing = experts.stream()
                    .filter(w -> meeting.getConfirmedRoster() || !w.getConfirmedRoster())
                    .collect(Collectors.toList());
            if (expertNoticing.isEmpty()) {
                return;
            }
            List<Long> currConfirmedMeIds = CollUtils.fieldList(expertNoticing, MeetingExpert::getId);
            LambdaUpdateWrapper<MeetingExpert> meUpdate = Wrappers.lambdaUpdate(MeetingExpert.class)
                    .in(MeetingExpert::getId, currConfirmedMeIds)
                    .set(MeetingExpert::getConfirmedRoster, Boolean.TRUE);
            meetingExpertService.update(meUpdate);
            meetingCallOrMsgHelper.sendConfirmedRosterMsg(expertNoticing, meeting);
        } finally {
            distributedLock.releaseLock(key);
        }
    }

    public PageVo<ReviewProjectDTO> pageReviewProject(ReviewProjectListReq req) {
        Page<ReviewProjectDTO> page = meetingExpertService.pageReviewProjectList(req);
        return PageVo.of(page.getRecords(), page.getTotal());
    }

    public PageVo<MeetingReviewProjectVO> optionProject(MeetingOptionProjectReq req) {
        String meetingType = req.getMeetingType();
        LambdaQueryWrapper<Project> query = Wrappers.lambdaQuery(Project.class);
        switch (MeetingReviewTypeEnum.getByCode(meetingType)) {
            case PRELIMINARY_SCHEME_REVIEW:
                buildOptionProjectQuery(query, meetingType, ProjectStatusEnum.PRE_APPLYING);
                break;
            case CONSTRUCTION_SCHEME_REVIEW:
                buildOptionProjectQuery(query, meetingType, ProjectStatusEnum.SCHEME_UNDER_REVIEW);
                break;
            case ACCEPTANCE_SCHEME_REVIEW:
                buildOptionProjectQuery(query, meetingType, ProjectStatusEnum.FINAL_ACCEPTANCE_IS_UNDER_REVIEW);
                break;
            case DEPT_JOIN_REVIEW:
                buildOptionProjectQuery(query, meetingType, ProjectStatusEnum.DEPARTMENT_JOINT_REVIEW);
                break;
            default:
                return PageVo.empty();
        }
        UserInfoDetails userDetail = LoginUserUtil.loginUserDetail();
        Collection<String> regionCodes = regionCache.listChildRegionCodeList(userDetail.getRegionCode(), userDetail.getRegionLevel());
        query.in(Project::getAreaCode, regionCodes);
        Page<Project> page = projectService.page(req.page(), query);
        PageVo<MeetingReviewProjectVO> result = PageVo.of(null, page.getTotal());
        if (result.getTotal() > 0) {
            result.setRecords(CollUtils.convert(page.getRecords(), this::buildProjectList));
        }
        return result;
    }

    public List<MeetingReviewProjectVO> projectsByMeetingId(Long meetingId) {
        Meeting meeting = meetingService.getById(meetingId);
        if (meeting.getIsInnerProject()) {
            List<MeetingInnerProject> projects = meetingInnerProjectService.listByMeetingId(meetingId);
            List<Long> projectIdList = CollUtils.fieldList(projects, MeetingInnerProject::getProjectId);
            List<Project> projectList = projectService.listByIds(projectIdList);
            Map<Long, ExpertReview> reviewMap = expertReviewService.listFinalReviewMap(meetingId);
            return CollUtils.convert(projectList, w -> {
                MeetingReviewProjectVO vo = buildProjectList(w);
                ExpertReview review = reviewMap.get(w.getId());
                if (review != null) {
                    vo.setReviewResult(ReviewResultEnum.getByCode(review.getReviewResult()).getValue());
                } else {
                    vo.setReviewResult("--");
                }
                return vo;
            });
        } else {
            List<MeetingOuterProject> projects = meetingOuterProjectService.listByMeetingId(meetingId);
            return BeanUtil.copyToList(projects, MeetingReviewProjectVO.class);
        }
    }

    private MeetingReviewProjectVO buildProjectList(Project project) {
        return MeetingReviewProjectVO
                .builder()
                .id(project.getId())
                .projectCode(project.getProjectCode())
                .projectName(project.getProjectName())
                .declaredAmount(project.getDeclareAmount())
                .projectType(project.getProjectType())
                .projectYear(project.getProjectYear())
                .buildOrg(project.getBuildOrgName())
                .createOn(project.getCreateOn())
                .build();
    }

    private void buildOptionProjectQuery(LambdaQueryWrapper<Project> query, String meetingType, ProjectStatusEnum status) {
        query.eq(Project::getStatus, status.getCode());
        String sql = String.format("select 1 from meeting m inner join meeting_inner_project mip on" +
                " m.is_inner_project = true and m.id = mip.meeting_id and nd_project.id = mip.project_id" +
                " and m.type = %s and m.status != 3", meetingType);
        String sql2 = String.format("select review_result in (2,3) from " +
                " (select ner.review_result from meeting m inner join meeting_inner_project mip on" +
                " m.is_inner_project = true and m.id = mip.meeting_id and nd_project.id = mip.project_id" +
                " and m.type = %s and m.status != 3 left join nd_expert_review ner on ner.meeting_id = m.id " +
                " and ner.project_id = mip.project_id and ner.is_final = true order by m.create_on desc limit 1) t", meetingType);
        query.and(q1 -> q1.notExists(sql).or(q2 -> q2.apply(sql2)));
    }

}
